/* String extension library
 * J. Welsh, May 2019
 * Public domain
 */

#include <string.h> /* strlen */
#include "strext.h"

int mem_eq(void const *va, void const *vb, size_t len) {
	uchar const *a = va, *b = vb;

	for (; len; --len, ++a, ++b)
		if (*a != *b) return 0;
	return 1;
}

int str_eq(char const *a, char const *b) {
	for (;; ++a, ++b) {
		if (!*a)
			return !*b;
		if (*a != *b)
			return 0;
	}
}

int ustr_eq_cstr(uchar const *a, char const *b) {
	for (;; ++a, ++b) {
		if (!*a)
			return !*b;
		if (*a != *b)
			return 0;
	}
}

int ustr_eq_cstr_l(uchar const *a, char const *b, size_t len) {
	for (; len; --len, ++a, ++b) {
		if (!*a)
			return !*b;
		if (*a != *b)
			return 0;
	}
	return 1;
}

size_t stristr(char const *s, char const *sub) {
	char const head = *sub, *tail = sub+1;
	size_t i, tlen;

	if (!head) return 0; /* empty string */
	tlen = strlen(tail);
	for (i=0; *s; ++i, ++s)
		if (*s == head && mem_eq(s+1,tail,tlen)) break;
	/* mem_eq won't overflow "s" because it short-circuits on non-match,
	 * and the only way for NUL to match tail is to have exceeded tlen
	 * (or if the contents were mutated asynchronously...) */
	return i;
}

size_t strichr(char const *s, char c) {
	size_t i;

	for (i=0; *s && *s != c; ++i, ++s) ;
	return i;
}

size_t strlcpy(char *d, char const *s, size_t len) {
	if (len) {
		size_t space = --len;

		for (; space; --space)
			if (!(*d++ = *s++)) return len-space;
		*d = 0;
	}
	return len+strlen(s);
}

size_t strlcat(char *d, char const *s, size_t len) {
	size_t dlen = 0;

	for (; dlen != len && *d; ++d) ++dlen;
	return dlen+strlcpy(d,s,len-dlen);
}

#ifdef TEST
#include <assert.h>
int main() {
	char d[4];

	assert(strlcpy(NULL,"",0) == 0);

	*d = 'a';
	assert(strlcpy(d,"",1) == 0); assert(!*d);

	assert(strlcpy(NULL,"a",0) == 1);
	assert(strlcpy(d,"a",1) == 1); assert(str_eq(d,""));
	assert(strlcpy(d,"a",2) == 1); assert(str_eq(d,"a"));
	assert(strlcpy(d,"a",3) == 1); assert(str_eq(d,"a"));

	assert(strlcpy(NULL,"ab",0) == 2);
	assert(strlcpy(d,"ab",1) == 2); assert(str_eq(d,""));
	assert(strlcpy(d,"ab",2) == 2); assert(str_eq(d,"a"));
	assert(strlcpy(d,"ab",3) == 2); assert(str_eq(d,"ab"));
	assert(strlcpy(d,"ab",4) == 2); assert(str_eq(d,"ab"));

	assert(strlcat(NULL,"",0) == 0);
	assert(strlcat(NULL,"a",0) == 1);

	strlcpy(d,"",sizeof d);
	assert(strlcat(d,"",1) == 0); assert(str_eq(d,""));
	assert(strlcat(d,"",2) == 0); assert(str_eq(d,""));

	strlcpy(d,"a",sizeof d);
	assert(strlcat(d,"",0) == 0); assert(str_eq(d,"a"));
	assert(strlcat(d,"",1) == 1); assert(str_eq(d,"a"));
	assert(strlcat(d,"",2) == 1); assert(str_eq(d,"a"));
	assert(strlcat(d,"b",0) == 1); assert(str_eq(d,"a"));
	assert(strlcat(d,"b",1) == 2); assert(str_eq(d,"a"));
	assert(strlcat(d,"b",2) == 2); assert(str_eq(d,"a"));
	assert(strlcat(d,"b",3) == 2); assert(str_eq(d,"ab"));

	strlcpy(d,"aa",sizeof d);
	assert(strlcat(d,"b",0) == 1); assert(str_eq(d,"aa"));
	assert(strlcat(d,"b",1) == 2); assert(str_eq(d,"aa"));
	assert(strlcat(d,"b",2) == 3); assert(str_eq(d,"aa"));
	assert(strlcat(d,"b",3) == 3); assert(str_eq(d,"aa"));
	assert(strlcat(d,"b",4) == 3); assert(str_eq(d,"aab"));

	return 0;
}
#endif
