#!/bin/sh
# Automated Gales Linux bootstrap: based on the BUILD document, with a few adjustments to make based on your system and kernel selection.
# J. Welsh, Dec 2020

# NOTE: for now this is to be kept in sync with the manual process in BUILD.

set -e

# Entry point to re-exec self in a clean environment
if [ -z "$GALES_BOOTSTRAP_STAGE" ] ; then
	GALES_REPO=$PWD # must be writeable only due to the final mkinstaller step
	if ! [ -f bootstrap.sh -a -f mkinstaller ] ; then
		echo "must be run from the top of the Gales repository" >&2
		exit 1
	fi
	cd /var/build # adjust as needed (must be writeable)
	rm -rf gales
	mkdir gales
	cd gales
	printf 'logging to %s and %s\n' "$PWD/bootstrap.log" "$PWD/checkpoint.log"
	exec env -i TERM=$TERM HOME=$HOME USER=$USER LC_ALL=C PATH=$PWD/tools/bin:$PATH GALES_BOOTSTRAP_STAGE=1 GALES_REPO=$GALES_REPO /bin/sh $GALES_REPO/bootstrap.sh >bootstrap.log 2>&1
fi

set -x

export GALES=$PWD
S=/gales/dist/src # adjust as needed
P=$GALES_REPO/base/patches
TOOLS=$PWD/tools
ROOT=$PWD/root
BUILD=x86_64-unknown-linux-gnu
TGT=x86_64-gales-linux-musl
J=-j1 # adjust as needed

exec 3>checkpoint.log
time_start=`date +%s`
# hmm, I included "stat" in the build-time busybox config so the following would work portably but that's not available at this stage... perhaps in the end I'm stuck scraping "df" anyway
free_blocks_start=`stat -f $GALES -c %f`
checkpoint () {
	# Positional parameters:
	# 1. Checkpoint name/number (filesystem friendly)
	# 2. Description
	# Presently all we do is log time (seconds) and filesystem usage (binary kB, assuming 4096 byte blocks) at what are likely to be the major local extrema.
	time=`date +%s`
	free_blocks=`stat -f $GALES -c %f`
	printf '%d %d %s %s\n' `expr $time - $time_start` `expr \( $free_blocks_start - $free_blocks \) \* 4` "$1" "$2" >&3
}

checkpoint 2.1 "Arrange in-tree sources into the build tree"
cd $GALES
mkdir build root
mkdir -m 1755 root/package
mkdir -m 1755 root/package/admin
cp -r $GALES_REPO/base/* build/
mv build/daemontools root/package/admin/daemontools-0.76

checkpoint 2.2 "Compile basic build-time tools using the host toolchain"
cd build/busybox
cp $GALES_REPO/kconfig/busybox-1.24-build-tools.config .config
make $J
mkdir -p $TOOLS/bin
mv busybox $TOOLS/bin/
$TOOLS/bin/busybox --install $TOOLS/bin

checkpoint 2.3 "Verify and unpack out-of-tree sources to the build tree"
cd $S
sha512sum -c $GALES_REPO/base/src.sha512

cd $GALES/build
tar xzf $S/bin86-0.16.17.tar.gz
tar xjf $S/binutils-2.24.tar.bz2
tar xJf $S/e2fsprogs-1.43.4.tar.xz
tar xjf $S/gcc-4.7.4.tar.bz2
tar xJf $S/gmp-6.0.0a.tar.xz
tar xzf $S/lilo-24.0.tar.gz
tar xJf $S/linux-4.9.tar.xz
tar xjf $S/make-4.2.1.tar.bz2
tar xzf $S/mpc-1.0.3.tar.gz
tar xJf $S/mpfr-3.1.5.tar.xz

#3. Build a cross toolchain and environment using the host toolchain
checkpoint 3.1 "Prepare toolchain sources"
mkdir build-{binutils,gcc}

cd binutils-2.24
rm -r intl gold include/xregex2.h libiberty/regex.c
patch -p1 < $P/binutils-2.24-system-regex.patch
patch -p1 < $P/binutils-2.24-ld-kill-sneaky-plugins.patch
patch -p1 < $P/binutils-2.24-bounded-stack.patch
patch -p1 < $P/binutils-2.24-format-nonliteral.patch
# Bypass spurious post-patch info generation to avoid makeinfo dependency.
# (This "hypertext must be precompiled" thing is dumb.)
(cd bfd/doc && touch archive.texi bfdt.texi bfd.texinfo bfd.info)
cd ..

cd gcc-4.7.4
rm -r boehm-gc gnattools intl libmudflap libssp libada libffi libgfortran libgo libjava libobjc lto-plugin libiberty include/xregex*
patch -p1 < $P/gcc-4.7.4-musl.patch
patch -p1 < $P/gcc-4.7.4-58385-fold-const-side-effect.patch
patch -p1 < $P/gcc-4.7.4-49008-59545-genautomata.patch
patch -p1 < $P/gcc-4.7.4-31798-arm-static-libs.patch
patch -p1 < $P/gcc-4.7.4-55003-constexpr-no-initializer.patch
patch -p1 < $P/gcc-4.7.4-kill-fixincludes.patch
patch -p1 < $P/gcc-4.7.4-hardening.patch
patch -p1 < $P/gcc-4.7.4-libgcc-bounded-stack.patch
patch -p1 < $P/gcc-4.7.4-libcpp-format-nonliteral.patch
patch -p1 < $P/gcc-4.7.4-demangler-amputation.patch
ln -s ../binutils-2.24/libiberty .
ln -s ../../binutils-2.24/include/xregex.h include/
ln -s ../gmp-6.0.0 gmp
ln -s ../mpfr-3.1.5 mpfr
ln -s ../mpc-1.0.3 mpc
cd ..

cd gmp-6.0.0
patch -p1 < $P/gmp-6.0.0-no-asm.patch
patch -p1 < $P/gmp-6.0.0-format-nonliteral.patch
cd ..

cd mpfr-3.1.5
patch -p1 < $P/mpfr-3.1.5-bounded-stack.patch
patch -p1 < $P/mpfr-3.1.5-format-nonliteral.patch
cd ..

checkpoint 3.2 "Build cross binutils"
cd build-binutils
CFLAGS="-std=gnu89 -O0" ../binutils-2.24/configure --build=$BUILD --host=$BUILD --target=$TGT --prefix=$TOOLS --with-sysroot=$ROOT --disable-{nls,shared,werror} --enable-deterministic-archives
make $J
make install
checkpoint 3.2.1 "cross binutils cleanup"
rm -rf *
cd ..

checkpoint 3.3 "Build a minimal C cross compiler for building libc"
cd build-gcc
CFLAGS="-std=gnu89 -O0" ../gcc-4.7.4/configure --build=$BUILD --host=$BUILD --target=$TGT --prefix=$TOOLS --with-sysroot=$ROOT --with-newlib --without-headers --with-local-prefix=$TOOLS --with-native-system-header-dir=/include --with-mpfr-include=$PWD/../gcc-4.7.4/mpfr/src --with-mpfr-lib=$PWD/mpfr/src/.libs --disable-{bootstrap,decimal-float,fixincludes,libgcc,libgomp,libmudflap,libquadmath,libssp,lto,multilib,nls,shared,threads,werror} --enable-languages=c
make $J
make install
checkpoint 3.3.1 "minimal cross gcc cleanup"
rm -r *
cd ..

checkpoint 3.4 "Cross compile musl libc, installed to the compiler's sysroot"
cd musl
./configure --build=$BUILD --host=$TGT --prefix= --disable-shared --enable-debug --enable-warnings CFLAGS=-O1
make $J install DESTDIR=$ROOT
cd ..

checkpoint 3.5 "Build the full cross compilers"
cd build-gcc
CFLAGS="-std=gnu89 -O2" ../gcc-4.7.4/configure --build=$BUILD --host=$BUILD --target=$TGT --prefix=$TOOLS --with-sysroot=$ROOT --with-local-prefix=$TOOLS --with-native-system-header-dir=/include --with-mpfr-include=$PWD/../gcc-4.7.4/mpfr/src --with-mpfr-lib=$PWD/mpfr/src/.libs --disable-{bootstrap,decimal-float,fixincludes,libgomp,libitm,libmudflap,libquadmath,libssp,libstdc++-v3,lto,multilib,nls,shared,werror} --enable-{threads=posix,__cxa_atexit} --enable-languages=c,c++
make $J # on my Core 2 this takes 15m user time
make install
checkpoint 3.5.1 "full cross gcc cleanup"
rm -r *
cd ..

checkpoint 3.6 "Install the kernel userspace API headers"
cd linux-4.9
patch -p1 < $P/linux-4.9-portable-build.patch
make headers_install
find usr/include -name '.*install*' -exec rm {} +
cp -r usr/include $ROOT
cd ..

checkpoint 4 "Cross compile the kernel"
cd linux-4.9
xz -dc /home/user/archive/src/linux/patch-4.9.95.xz | patch -p1 # adjust as needed
cp $GALES_REPO/kconfig/linux-4.9-x86_64-thinkpad-static.config .config # adjust as needed

rm -f .version
make KBUILD_BUILD_TIMESTAMP=@0 KBUILD_BUILD_USER=build KBUILD_BUILD_HOST=localhost CROSS_COMPILE=$TGT- $J
cp arch/x86/boot/bzImage $GALES/bzImage-4.9.95 # adjust version as needed for your own reference

cd ..

#5. Cross compile the base userland
checkpoint 5.1 "BusyBox"
cd busybox
make clean
cp $GALES_REPO/kconfig/busybox-1.24.config .config
make CROSS_COMPILE=$TGT- $J
mkdir -p $ROOT/bin
mv busybox $ROOT/bin/
cd ..

checkpoint 5.2 "Non-BusyBox Unix utilities"
cd gksh
make CROSS_COMPILE=$TGT- $J
DEST=$ROOT/gales/pkg/gksh-0.0-99K
mkdir -p $DEST/{bin,man/man1}
cp ksh $DEST/bin/
cp ksh.1 $DEST/man/man1/
cd ..

cd make-4.2.1
CFLAGS=-O1 ./configure --build=$BUILD --host=$TGT --prefix=/gales/pkg/make-0.0-4.2.1 --datarootdir='${prefix}' --disable-nls
make $J
make install DESTDIR=$ROOT
cd ..

checkpoint 5.3 "Linux specific utilities"
cd makedev
make install CC=$TGT-gcc DESTDIR=$ROOT sbindir=/gales/pkg/MAKEDEV-0.0-3.23/bin mandir=/gales/pkg/MAKEDEV-0.0-3.23/man
cd ..

cd e2fsprogs-1.43.4
patch -p1 < $P/e2fsprogs-1.43.4-really-unbreak-cross.patch
rm -r doc
CFLAGS=-O1 ./configure --build=$BUILD --host=$TGT --prefix=/gales/pkg/e2fsprogs-0.0-1.43.4 --datarootdir='${prefix}' --sysconfdir=/etc --disable-{e2initrd-helper,nls,uuidd} --enable-{libuuid,libblkid,symlink-install}
make $J install DESTDIR=$ROOT

cd $ROOT/etc
mkdir examples
mv makedev.d examples/
sed -e '/enable_periodic_fsck/ s/0/1/' mke2fs.conf > examples/mke2fs.conf
rm mke2fs.conf
cd $GALES/build

checkpoint 5.4 "Init and process supervision system"
cd init
make install CC=$TGT-gcc DESTDIR=$ROOT

cd $ROOT/package/admin/daemontools-0.76
CHOST=$TGT- package/install
rm -rf compile
cd $GALES/build

checkpoint 5.5 "Bootloader"
cd bin86-0.16.17
patch -p1 < $P/bin86-0.16.17-int32.patch
patch -p1 < $P/bin86-0.16.17-install.patch
make $J PREFIX=$TOOLS CFLAGS=-D_POSIX_SOURCE install
cd ..

cd lilo-24.0
patch -p1 < $P/lilo-24.0-musl-v2.patch
patch -p1 < $P/lilo-24.0-cross.patch
patch -p1 < $P/lilo-24.0-kill-format-juggling-and-devfs.patch
patch -p1 < $P/lilo-24.0-blkext.patch
cd src
make all CROSS_COMPILE=$TGT- $J
DEST=$ROOT/gales/pkg/lilo-0.0-24.0
mkdir -p $DEST/{bin,man/man{5,8}}
cp lilo $DEST/bin/
cd ../man
cp lilo.8 $DEST/man/man8/
cp lilo.conf.5 $DEST/man/man5/
cp $GALES_REPO/conf/lilo*.conf $ROOT/etc/examples/
cd ../..

checkpoint 6 "Cross compile a native toolchain in the sysroot"
cd build-binutils
../binutils-2.24/configure --build=$BUILD --host=$TGT --target=$TGT --prefix=/gales/pkg/binutils-0.0-2.24 --disable-{nls,shared,werror} --enable-deterministic-archives
make $J
STRIPPROG=$TGT-strip make install-strip DESTDIR=$ROOT
checkpoint 6.1 "native binutils cleanup"
rm -rf *
cd ..

checkpoint 6.2 "native gcc build"
cd build-gcc
../gcc-4.7.4/configure --build=$BUILD --host=$TGT --target=$TGT --prefix=/gales/pkg/gcc-0.0-4.7.4 --with-local-prefix=/local --with-native-system-header-dir=/include --with-mpfr-include=$PWD/../gcc-4.7.4/mpfr/src --with-mpfr-lib=$PWD/mpfr/src/.libs --disable-{bootstrap,fixincludes,libgomp,libmudflap,libssp,libstdcxx-pch,lto,multilib,nls,shared,werror} --enable-{threads=posix,__cxa_atexit} --enable-languages=c,c++
make $J
make install-strip DESTDIR=$ROOT
checkpoint 6.3 "native gcc cleanup"
rm -r *
cd ..

cd gcc-4.7.4/gcc/doc
DEST=$ROOT/gales/pkg/gcc-0.0-4.7.4
mkdir -p $DEST/{info,man/man1}
cp {cpp,cppinternals,gcc,gccinstall,gccint}.info* $DEST/info/
cp {cpp,g++,gcc,gcov}.1 $DEST/man/man1/
cd $DEST
mv share/info/lib{itm,quadmath}.info* info/

rm -r share/{info,man}
rm lib*/*.la
# ^ there used to be a libiberty.a to remove here, apparently no longer installed but unsure why
rm bin/*gcc-*
rm libexec/gcc/*/*/lto-wrapper

cd ../binutils-0.0-2.24
mv share/{info,man} .
rmdir share
rm -r lib

checkpoint 7 "Make symlinks to activate the installed versions"
cd $ROOT/gales
mkdir command info man
cd pkg
for p in MAKEDEV binutils e2fsprogs gcc gksh init lilo make ; do ln -s $p-*-* $p ; done
cd ../command
ln -s ../pkg/{MAKEDEV,binutils,e2fsprogs,gcc,lilo,make}/bin/* .
ln -s ../pkg/e2fsprogs/sbin/* .
ln -s ../pkg/gcc/bin/gcc cc
cd ../info
ln -s ../pkg/{binutils,gcc,make}/info/*.info* .
cd ../man
for p in MAKEDEV binutils e2fsprogs gcc gksh init lilo make ; do ln -s ../pkg/$p/man $p ; done
cd $ROOT/bin
ln -s ../gales/pkg/init/bin/init .
ln -s ../gales/pkg/gksh/bin/ksh .
ln -s ../gales/pkg/gksh/bin/ksh sh

checkpoint 8 "Pack distribution archives"
cd $GALES
$TGT-strip root/bin/* root/gales/command/*
export DATE=`date +%Y%m%d`
tar -czf base-$DATE.tar.gz --sterilize -C root .
cd $GALES_REPO
./mkskeleton > $GALES/skeleton-$DATE.sh
./mkinstaller

checkpoint 9 "Create a bootable install medium, if desired"
